import socket
import select
import platform
import multiprocessing
import ctypes
import threading
import collections
import queue
import errno
import json
import traceback
from datetime import datetime

from contextlib import contextmanager


@contextmanager
def __acquire_timeout(self, lock:threading.Lock, timeout:float):
    result = lock.acquire(timeout=timeout)
    try:
        yield result
    finally:
        if result:
            lock.release()
@contextmanager
def __acquire_blocking(self, lock:threading.Lock, blocking:bool):
    result = lock.acquire(blocking=blocking)
    try:
        yield result
    finally:
        if result:
            lock.release()


class KqueueServer:
    def __init__(self) -> None:
        self.__buffer_size = 10240
        self.__is_running = multiprocessing.Value(ctypes.c_bool, False)
        self.client_by_fileno = collections.defaultdict(dict)
        self.__kevent_by_fileno = collections.defaultdict(select.kevent)
        self.__recv_queue = queue.Queue()
        
        self.__detect_epollin_fileno_queue = queue.Queue()
        self.__send_fileno_queue = queue.Queue()
    
    def start(self, listen_ip:str, listen_port:int, is_blocking:bool = False, backlog:int = 5):
        self.__listen_socket = self.get_listener(listen_ip, listen_port, is_blocking, backlog)

        self.__is_running.value = True
        
        self.__kq_thread = threading.Thread(target=self.__kq_thread_function)
        self.__recv_work_thread = threading.Thread(target=self.__recv_work)
        self.__send_work_thread = threading.Thread(target=self.__send_work)
        
        self.__kq_thread.start()
        self.__recv_work_thread.start()
        self.__send_work_thread.start()
        
        listen_socket_fileno = self.__listen_socket.fileno()
        kevent = select.kevent(listen_socket_fileno)
        self.__kevent_by_fileno[listen_socket_fileno] = kevent
        
    def stop(self):
        self.__is_running.value = False
        self.__listen_socket.shutdown(socket.SHUT_RDWR)
        
    def join(self):
        self.__kq_thread.join()
        self.__recv_work_thread.join()
        self.__send_work_thread.join()
    
    def create_client(self, client_socket):
        return {
            "socket" : client_socket,
            "lock" : threading.Lock(),
            "send_buffer_queue" : queue.Queue(),
            "sending_buffer" : b''
        }
    
    def recv(self):
        return self.__recv_queue.get()
    
    def __recv_work(self):
        while self.__is_running.value:
            detect_fileno = self.__detect_epollin_fileno_queue.get()
            if not detect_fileno:
                self.__recv_queue.put_nowait(None)
                break
            
            result = b''
            with __acquire_timeout(self.client_by_fileno[detect_fileno]['lock'], 1) as acqiured:
                try:
                    while True:
                        recv_bytes = self.client_by_fileno[detect_fileno]['socket'].recv(self.__buffer_size)
                        if recv_bytes == None or recv_bytes == -1 or recv_bytes == b'':
                            break
                        elif recv_bytes:
                            result += recv_bytes
                        else:
                            break
                except BlockingIOError as e:
                    if e.errno == socket.EAGAIN:
                        pass
                    else:
                        raise e
                if result is not None and result != b'':
                    self.__recv_queue.put_nowait({
                        "fileno": detect_fileno,
                        "data": result
                    })

#####################################################################################################################
#####################################################################################################################
#####################################################################################################################
    def send(self, fileno:int, data:bytes):
        self.client_by_fileno[fileno]['send_buffer_queue'].put_nowait(data)
        self.__send_fileno_queue.put_nowait(fileno)
    
    def __send_work(self):
        while self.__is_running.value:
            send_fileno = self.__send_fileno_queue.get()
            if not send_fileno:
                break
        
            client_data = self.client_by_fileno.get(send_fileno)
            if client_data:
                with __acquire_timeout(self.client_by_fileno[send_fileno]['lock'], 1) as acqiured:
                    if acqiured:
                        if self.client_by_fileno[send_fileno]['sending_buffer'] == b'':
                            try:
                                self.client_by_fileno[send_fileno]['sending_buffer'] = self.client_by_fileno[send_fileno]['send_buffer_queue'].get_nowait()
                                self.client_by_fileno[send_fileno]['socket'].send(b'')
                            except queue.Empty:
                                return
                        
                        start_index = 0
                        end_index = len(self.client_by_fileno[send_fileno]['sending_buffer'])
                        try:
                            while start_index < end_index:
                                send_length = self.client_by_fileno[send_fileno]['socket'].send(self.client_by_fileno[send_fileno]['sending_buffer'][start_index:end_index])
                                if send_length <= 0:
                                    break
                                start_index += send_length
                        except BlockingIOError as e:
                            if e.errno == socket.EAGAIN:
                                pass
                            else:
                                raise e
                            
                        if 0 <= start_index < end_index:
                            self.client_by_fileno[send_fileno]['sending_buffer'] = self.client_by_fileno[send_fileno]['sending_buffer'][start_index:end_index]
                        else:
                            self.client_by_fileno[send_fileno]['sending_buffer'] = b''
                            self.client_by_fileno[send_fileno]['socket'].send(b'')
                        
                        if self.client_by_fileno[send_fileno]['sending_buffer'] != b'':
                            self.__send_fileno_queue.put_nowait(send_fileno)
                        elif not self.client_by_fileno[send_fileno]['send_buffer_queue'].empty():
                            self.__send_fileno_queue.put_nowait(send_fileno)
                        
                    # else:
                        # print('send lock timeout')
                            
#####################################################################################################################
#####################################################################################################################
#####################################################################################################################            
            
    
    def get_listener(self, listen_ip:str, listen_port:int, is_blocking:bool = False, backlog:int = 5) -> socket.socket:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.setblocking(is_blocking)
        s.bind((listen_ip, listen_port))
        s.listen(backlog)
        return s 

    def __kq_thread_function(self):
        kq = select.kqueue()
        while self.__is_running.value:
            kevents = self.__kevent_by_fileno.values()
            events = kq.control(list(kevents), 1000)
            for event in events:
                if event.flags & select.KQ_EV_ERROR:
                    # print("event.flags & select.KQ_EV_ERROR")
                    return
                    
                if event.ident == self.__listen_socket.fileno():
                    if event.flags & select.KQ_EV_EOF:
                        # print(f"event.ident == detect_close_fd.fileno() {event}")
                        self.__is_running.value = False
                        continue
                    else:
                        client_socket, address = self.__listen_socket.accept()
                        client_socket_fileno = client_socket.fileno()
                        # print(f"accept {client_socket_fileno} {address}")
                        client_socket.setblocking(False)
                        client = self.client_by_fileno.get(client_socket_fileno)
                        if client:
                            s:socket.socket = client["socket"]
                            try:
                                s.shutdown(socket.SHUT_RDWR)
                            except Exception as e:
                                pass
                            s.close()
                        client_data = self.create_client(client_socket)
                        self.client_by_fileno.update({client_socket_fileno : client_data})
                        
                        kevent = select.kevent(client_socket_fileno)
                        self.__kevent_by_fileno.update({client_socket_fileno:kevent})
                        
                elif event.filter == select.KQ_FILTER_READ:
                    if event.flags & select.KQ_EV_EOF:
                        # print("event.flags & select.KQ_EV_EOF")
                        self.__kevent_by_fileno.pop(event.ident)
                        client_data = self.client_by_fileno.pop(event.ident)
                        client_data['socket'].shutdown(socket.SHUT_RDWR)
                        client_data['socket'].close()
                
                    else:
                        client_data = self.client_by_fileno.pop(event.ident)
                        data = client_data['socket'].recv()
                        # print(data)
                
        kq.close()
        
        
        




import socket
import errno

class DarwinClient:
    def __init__(self) -> None:
        self.__buffer_size = 8196
        self.__client_socket : socket.socket = None
        
    def connect(self, ip:str, port:int):
        self.__client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__client_socket.connect((ip, port))
        # self.__client_socket.setblocking(False)
    
    def shutdown(self):
        if self.__client_socket:
            try:
                self.__client_socket.shutdown(socket.SHUT_RDWR)
            except OSError as e:
                if e.errno == errno.ENOTCONN:
                    pass
                else:
                    raise e
        
    def close(self):
        self.__client_socket.close()
        self.__client_socket = None
            
    def sendall(self, data:bytes):
        self.__client_socket.sendall(data)
    
    def send(self, data:bytes) -> int:
        return self.__client_socket.send(data)
        
    def recv(self, buffer_size:int = None) -> bytes:
        if buffer_size:
            return self.__client_socket.recv(buffer_size)
        else:
            return self.__client_socket.recv(self.__buffer_size)
        
    def fileno(self) -> int:
        return self.__client_socket.fileno()
    
    def setblocking(self, is_block:bool):
        self.__client_socket.setblocking(is_block)



import socket
import select
import multiprocessing
import ctypes
import threading
import collections
import queue
import errno
import traceback
from datetime import datetime

class RelayServer():
    def __init__(self) -> None:
        self.__buffer_size = 8196
        self.__is_running = multiprocessing.Value(ctypes.c_bool, False)
        self.__message_queue = None
        
        self.__running_threads = []
        self.__running_thread_by_tid = collections.defaultdict(threading.Thread)
        
        self.__listener_by_ip_port = collections.defaultdict(socket.socket)
        self.__listener_by_fileno = collections.defaultdict(socket.socket)
        
        self.__socket_by_fileno = collections.defaultdict(socket.socket)
        
        self.__client_by_fileno = collections.defaultdict(socket.socket)
        self.__listener_fileno_by_client_fileno = collections.defaultdict(int)
        self.__client_fileno_dict_by_listener_fileno = collections.defaultdict(dict)
        self.__registered_eventmask_by_fileno = collections.defaultdict(int)
        
        self.__send_lock_by_fileno = collections.defaultdict(threading.Lock)
        self.__recv_lock_by_fileno = collections.defaultdict(threading.Lock)
        
        self.__send_buffer_by_fileno = collections.defaultdict(bytes)
        self.__send_buffer_lock_by_fileno = collections.defaultdict(threading.Lock)
        
        self.__relay_addr_by_listener_addr = collections.defaultdict(str)
        self.__relay_by_fileno = collections.defaultdict(socket.socket)
        self.__client_by_relay_fileno = collections.defaultdict(socket.socket)
        self.__relay_by_client_fileno = collections.defaultdict(socket.socket)
        
        self.__epoll = None
        
        self.__listener_eventmask = select.EPOLLIN | select.EPOLLPRI | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
        self.__recv_eventmask = select.EPOLLIN  | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
        self.__send_recv_eventmask = select.EPOLLIN | select.EPOLLOUT | select.EPOLLHUP | select.EPOLLRDHUP
        self.__closer_eventmask = select.EPOLLIN | select.EPOLLPRI | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
        
    
    def __listen(self, ip:str, port:int, backlog:int = 5):
        listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1) # Nagle's
        
        # increase buffer size
        recv_buf_size = listener.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
        send_buf_size = listener.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, recv_buf_size*2)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, send_buf_size*2)
        
        listener.setblocking(False)
        listener.bind((ip, port))
        listener.listen(backlog)
        
        listener_fileno = listener.fileno()
        
        self.__listener_by_ip_port.update({f"{ip}:{port}":listener})
        self.__listener_by_fileno.update({listener_fileno : listener})
        
        self.__client_fileno_dict_by_listener_fileno.update({listener_fileno : {}})
        
        if self.__epoll and  self.__epoll.closed:
            # After 'start()'
            self.__epoll.register(listener_fileno, self.__listener_eventmask)
            self.__registered_eventmask_by_fileno.update({listener_fileno : self.__listener_eventmask})

    def start(self, count_threads:int=1, message_queue = None ):
        self.__is_running.value = True
        self.__message_queue = message_queue
        
        self.__epoll = select.epoll()
        self.__close_event, self.__close_event_listener = socket.socketpair()
        self.__epoll.register(self.__close_event_listener, self.__closer_eventmask)
        
        for _ in range(count_threads):
            et = threading.Thread(target=self.__epoll_thread_function)
            et.start()
            self.__running_threads.append(et)
            self.__running_thread_by_tid[et.ident] = et
            
        for fileno in self.__listener_by_fileno:
            if fileno in self.__registered_eventmask_by_fileno:
                if self.__registered_eventmask_by_fileno[fileno] != self.__listener_eventmask:
                    self.__epoll.modify(fileno, self.__listener_eventmask)
            else:
                # After 'listen()'
                self.__epoll.register(fileno, self.__listener_eventmask)
                self.__registered_eventmask_by_fileno.update({fileno : self.__listener_eventmask})

    def send(self, socket_fileno:int, data:bytes):
        if data is not None and data != b'':
            try:
                send_buffer_lock = self.__send_buffer_lock_by_fileno.get(socket_fileno)
                if send_buffer_lock:
                    with send_buffer_lock:
                        send_buffer = self.__send_buffer_by_fileno.get(socket_fileno)
                        if send_buffer is not None:
                            self.__send_buffer_by_fileno.update({socket_fileno : send_buffer + data})
                    
                send_lock = self.__send_lock_by_fileno.get(socket_fileno)
                if send_lock:
                    with send_lock:
                        registered_eventmask = self.__registered_eventmask_by_fileno.get(socket_fileno)
                        if registered_eventmask is not None and registered_eventmask != self.__send_recv_eventmask:
                            self.__registered_eventmask_by_fileno.update({socket_fileno : self.__send_recv_eventmask})
                            if isinstance(self.__epoll, select.epoll):
                                self.__epoll.modify(socket_fileno, self.__send_recv_eventmask)
                
            except KeyError:
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send KeyError.\n{traceback.format_exc()}"
                    })
                
            except FileNotFoundError:
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send FileNotFoundError.\n{traceback.format_exc()}"
                    })
                    
            except PermissionError:
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send PermissionError.\n{traceback.format_exc()}"
                    })
                
            except OSError as e:
                if e.errno == errno.EBADF:
                    pass
                else:
                    raise e
            
    
    def join(self):
        for t in self.__running_threads:
            t:threading.Thread = t
            t.join()
                
    def close(self):
        self.__is_running.value = False
        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
            self.__message_queue.put_nowait(None)
        self.__shutdown_listeners()
        
        for _ in self.__running_threads:
            self.__close_event.send(b'close')
            tid_bytes = self.__close_event.recv(32)
            tid = int.from_bytes(tid_bytes, byteorder='big')
            self.__running_thread_by_tid[tid].join()
            
    def __shutdown_listeners(self):
        fileno_list = list(self.__listener_by_fileno.keys())
        for fileno in fileno_list:
            self.__shutdown_listener(fileno)
            
    def __shutdown_listener(self, listener_fileno:int):
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            listener.shutdown(socket.SHUT_RDWR)
        
    def __close_listener(self, listener_fileno:int):
        try:
            if isinstance(self.__epoll, select.epoll):
                self.__epoll.unregister(listener_fileno)
        except FileNotFoundError:
            pass
        except OSError as e:
            if e.errno == errno.EBADF:
                pass
            else:
                raise e
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            listener.close()
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{listener_fileno:3}] Listner Close()"
                })
        
    def __remove_listener(self, listener_fileno:int):
        try:
            _ = self.__listener_by_fileno.pop(listener_fileno)
        except KeyError:
            pass
    
    def __unregister(self, socket_fileno:int) -> bool:
        result = False
        try:
            _ = self.__registered_eventmask_by_fileno.pop(socket_fileno)
            if isinstance(self.__epoll, select.epoll):
                self.__epoll.unregister(socket_fileno)
            result = True
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{socket_fileno:3}] epoll unregister."
                })

        except KeyError:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{socket_fileno}] __unregister KeyError.\n{traceback.format_exc()}"
                })
            
        except FileNotFoundError:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{socket_fileno}] __unregister FileNotFoundError.\n{traceback.format_exc()}"
                })
            
        except OSError as e:
            if e.errno == errno.EBADF:
                pass
            else:
                raise e
            
        except Exception as e:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{socket_fileno}] send Exception: {e}.\n{traceback.format_exc()}"
                })
        
        return result
        
    def __shutdown_clients_by_listener(self, listener_fileno:int):
        client_fileno_dict = self.__client_fileno_dict_by_listener_fileno.get(listener_fileno)
        if client_fileno_dict:
            for client_fileno in client_fileno_dict:
                self.__shutdown_client(client_fileno)
        
    def __shutdown_client(self, socket_fileno:int):
        _socket = self.__socket_by_fileno.get(socket_fileno)
        if _socket:
            try:
                _socket.shutdown(socket.SHUT_RDWR)
            except ConnectionError:
                pass
            except OSError as e:
                if e.errno == errno.ENOTCONN: # errno 107
                    pass
                else:
                    raise e
            except Exception as e:
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send Exception: {e}.\n{traceback.format_exc()}"
                    })
            
    
    def __close_client(self, client_fileno:int):
        client_socket = self.__client_by_fileno.get(client_fileno)
        if client_socket:
            client_socket.close()
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{client_fileno:3}] Client Closed."
                })

    def __remove_client(self, client_fileno:int):
        try: _ = self.__send_lock_by_fileno.pop(client_fileno)
        except KeyError: pass
        try: _ = self.__recv_lock_by_fileno.pop(client_fileno)
        except KeyError: pass
        try: _ = self.__client_by_fileno.pop(client_fileno)
        except KeyError: pass
        
        len_remain_send_buffer = 0
        try: 
            remain_send_buffer = self.__send_buffer_by_fileno.pop(client_fileno)
            len_remain_send_buffer = len(remain_send_buffer)
        except KeyError: pass
        
        try:
            listener_fileno = self.__listener_fileno_by_client_fileno.pop(client_fileno)
            _ = self.__client_fileno_dict_by_listener_fileno[listener_fileno].pop(client_fileno)
        except KeyError: pass
        
        if 0 < len_remain_send_buffer:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{client_fileno}] Try Close. send buffer remain:{len_remain_send_buffer} bytes."
                })
        
    def __epoll_accept(self, listener_fileno:int):
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            client_socket = None
            try:
                client_socket, address = listener.accept()
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"accept {client_socket.fileno():3}:{address}"
                    })
                
                client_socket_fileno = client_socket.fileno()
                client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                client_socket.setblocking(False)
                
                self.__socket_by_fileno.update({client_socket_fileno : client_socket})
                self.__client_by_fileno.update({client_socket_fileno : client_socket})
                self.__send_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__recv_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__send_buffer_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__send_buffer_by_fileno.update({client_socket_fileno : b''})
                if not listener_fileno in self.__client_fileno_dict_by_listener_fileno:
                    self.__client_fileno_dict_by_listener_fileno.update({listener_fileno : {}})
                self.__client_fileno_dict_by_listener_fileno[listener_fileno][client_socket_fileno] = True
                self.__listener_fileno_by_client_fileno.update({client_socket_fileno : listener_fileno})
                
                self.__registered_eventmask_by_fileno.update({client_socket_fileno : self.__recv_eventmask})
                if isinstance(self.__epoll, select.epoll):
                    self.__epoll.register(client_socket, self.__recv_eventmask)
            except BlockingIOError as e:
                if e.errno == socket.EAGAIN:
                    return
                else:
                    raise e
            
            if client_socket:
                sockname = listener.getsockname()
                listener_addr = sockname[0]
                listener_port = sockname[1]
                
                disconnect_client = False
                
                relay_addr = self.__relay_addr_by_listener_addr.get(f"{listener_addr}:{listener_port}")
                if relay_addr:
                    relay_addrs = relay_addr.split(':')
                    relay_ip = relay_addrs[0]
                    relay_port = int(relay_addrs[1])
                    try:
                        relay_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        relay_socket.connect((relay_ip, relay_port))
                        relay_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        relay_socket.setblocking(False)
                        relay_socket_fileno = relay_socket.fileno()
                        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                            self.__message_queue.put_nowait({
                                "type" : "debug",
                                "message" : f"Client fileno:{client_socket.fileno()} -> Relay fileno:{relay_socket_fileno}"
                            })
                            
                        self.__registered_eventmask_by_fileno.update({relay_socket_fileno : self.__recv_eventmask})
                        if isinstance(self.__epoll, select.epoll):
                            self.__epoll.register(relay_socket_fileno, self.__recv_eventmask)
                        
                        self.__socket_by_fileno.update({relay_socket_fileno : relay_socket})
                        self.__relay_by_fileno.update({relay_socket_fileno : relay_socket})
                        self.__client_by_relay_fileno.update({relay_socket_fileno : client_socket})
                        self.__relay_by_client_fileno.update({client_socket_fileno : relay_socket})
                        
                        self.__send_lock_by_fileno.update({relay_socket_fileno : threading.Lock()})
                        self.__recv_lock_by_fileno.update({relay_socket_fileno : threading.Lock()})
                        self.__send_buffer_lock_by_fileno.update({relay_socket_fileno : threading.Lock()})
                        self.__send_buffer_by_fileno.update({relay_socket_fileno : b''})
                    except socket.error as e:
                        disconnect_client = True
                else:
                    disconnect_client = True
                
                if disconnect_client:
                    # client_socket.send(b'NOTSERVER')
                    self.__shutdown_client(client_socket.fileno())
                    
    def relay(self, fromip:str, fromport:int, toip:str, toport:int, check_relay_function=None):
        self.__listen(fromip, fromport)
        self.check_relay_function = check_relay_function
        self.__relay_addr_by_listener_addr.update({f"{fromip}:{fromport}" : f"{toip}:{toport}"})
    
    def __epoll_recv(self, recv_socket:socket.socket) -> bytes:
        recv_bytes = b''
        recv_socket_fileno = recv_socket.fileno()
        recv_lock = self.__recv_lock_by_fileno.get(recv_socket_fileno)
        if recv_lock:
            with recv_lock:
                if recv_socket:
                    is_connect = True
                    is_eagain = False
                    try:
                        temp_recv_bytes = recv_socket.recv(self.__buffer_size)
                        if temp_recv_bytes == None or temp_recv_bytes == -1 or temp_recv_bytes == b'':
                            is_connect = False
                        else:
                            recv_bytes += temp_recv_bytes
                            
                    except ConnectionError as e:
                        pass
                    except OSError as e:
                        if e.errno == socket.EAGAIN:
                            is_eagain = True
                        elif e.errno == errno.EBADF:
                            is_connect = False
                        else:
                            raise e

                    if not is_eagain and is_connect:
                        try:
                            if isinstance(self.__epoll, select.epoll):
                                self.__epoll.modify(recv_socket_fileno, self.__registered_eventmask_by_fileno[recv_socket_fileno])
                        except FileNotFoundError:
                            pass
                        except OSError as e:
                            if e.errno == errno.EBADF:
                                pass

        return recv_bytes
    
    def __epoll_send(self, client_fileno:int):
        is_connect = True
        try:
            client_socket = self.__socket_by_fileno.get(client_fileno)
            send_lock = self.__send_lock_by_fileno.get(client_fileno)
            send_buffer_lock = self.__send_buffer_lock_by_fileno.get(client_fileno)
            
            if send_lock is not None and client_socket is not None:
                with send_lock:
                    send_buffer = b''
                    if send_buffer_lock:
                        with send_buffer_lock:
                            send_buffer = self.__send_buffer_by_fileno.get(client_fileno)
                    else:
                        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                            self.__message_queue.put_nowait({
                                "type" : "debug",
                                "message" : f"[{client_fileno}] is_connect = False send_buffer_lock is {send_buffer_lock}"
                            })
                    
                    if send_buffer is not None:
                        if send_buffer != b'':
                            try:
                                sent_length = client_socket.send(send_buffer)
                                if 0 < sent_length:
                                    if send_buffer_lock:
                                        with send_buffer_lock:
                                            self.__send_buffer_by_fileno.update({client_fileno : send_buffer[sent_length:]})
                                    else:
                                        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                                            self.__message_queue.put_nowait({
                                                "type" : "debug",
                                                "message" : f"[{client_fileno}] is_connect = False send_buffer_lock is {send_buffer_lock}"
                                            })
                        
                            except ConnectionError as e:
                                pass
                            
                            except BlockingIOError as e:
                                if e.errno == socket.EAGAIN:
                                    pass
                                else:
                                    raise e
                                
                            except OSError as e:
                                if e.errno == errno.EBADF:
                                    is_connect = False
                                    
                                else:
                                    raise e
                    else:
                        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                            self.__message_queue.put_nowait({
                                "type" : "debug",
                                "message" : f"[{client_fileno}] is_connect = False send_buffer is {send_buffer}"
                            })
                        
                        is_connect = False

            else:
                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                    self.__message_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{client_fileno}] is_connect = False send_lock is {send_lock} and client_socket is {client_socket}"
                    })
                is_connect = False
        
            try:
                remain_buffer = b''
                send_buffer_lock = self.__send_buffer_lock_by_fileno.get(client_fileno)
                if send_buffer_lock:
                    with send_buffer_lock:
                        remain_buffer = self.__send_buffer_by_fileno.get(client_fileno)
                
                registered_eventmask = self.__registered_eventmask_by_fileno.get(client_fileno)
                if (remain_buffer is not None and registered_eventmask is not None) and\
                    remain_buffer == b'' and\
                    registered_eventmask != self.__recv_eventmask:
                    self.__registered_eventmask_by_fileno.update({client_fileno : self.__recv_eventmask})
                    if isinstance(self.__epoll, select.epoll):
                        self.__epoll.modify(client_fileno, self.__recv_eventmask)
            
            except OSError as e:
                if e.errno == errno.EBADF:
                    is_connect = False
                        
        except Exception as e:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{client_fileno}] __epoll_send Excetion:{e}.\n{traceback.format_exc()}"
                })
            
            is_connect = False
            
        return is_connect
    
    def __epoll_thread_function(self):
        __is_running = True
        tid = threading.get_ident()
        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
            self.__message_queue.put_nowait({
                "type" : "debug",
                "message" : f"Start Epoll Work"
            })
        
        try:
            if isinstance(self.__epoll, select.epoll):
                while __is_running:
                    events = self.__epoll.poll()
                    for detect_fileno, detect_event in events:
                        if detect_event & select.EPOLLPRI:
                            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                                self.__message_queue.put_nowait({
                                    "type" : "debug",
                                    "message" : f"EPOLLPRI [{detect_event:#06x} & select.EPOLLPRI]"
                                })
                            
                        if detect_fileno == self.__close_event_listener.fileno():
                            self.__close_event_listener.send(tid.to_bytes(32, 'big'))
                            __is_running = False
                            
                        elif detect_fileno in self.__listener_by_fileno:
                            if detect_event & (select.EPOLLHUP | select.EPOLLRDHUP):
                                self.__shutdown_clients_by_listener(detect_fileno)
                                if self.__unregister(detect_fileno):
                                    self.__close_listener(detect_fileno)
                                    self.__remove_listener(detect_fileno)
                                
                            elif detect_event & select.EPOLLIN:
                                self.__epoll_accept(detect_fileno)
                            
                            else:
                                if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                                    self.__message_queue.put_nowait({
                                        "type" : "debug",
                                        "message" : f"listen event else [{detect_event:#06x}]..?"
                                    })
                                
                        else:
                            if detect_event & select.EPOLLOUT:
                                if self.__epoll_send(detect_fileno) == False:
                                    if self.__unregister(detect_fileno):
                                        self.__close_client(detect_fileno)
                                        self.__remove_client(detect_fileno)
                            
                            if detect_event & select.EPOLLIN:
                                from_socket = None
                                to_socket = None
                                if detect_fileno in self.__client_by_fileno:
                                    from_socket = self.__client_by_fileno.get(detect_fileno)
                                    to_socket = self.__relay_by_client_fileno.get(detect_fileno)
                                    
                                elif detect_fileno in self.__relay_by_fileno:
                                    from_socket = self.__relay_by_fileno.get(detect_fileno)
                                    to_socket = self.__client_by_relay_fileno.get(detect_fileno)
                                
                                if from_socket and to_socket:
                                    recv_bytes = self.__epoll_recv(from_socket)
                                    if recv_bytes:
                                        if self.check_relay_function:
                                            if self.check_relay_function(from_socket.fileno(), to_socket.fileno(), recv_bytes):
                                                self.send(to_socket.fileno(), recv_bytes)
                                        else:
                                            self.send(to_socket.fileno(), recv_bytes)
                                        
                            if detect_event & (select.EPOLLHUP | select.EPOLLRDHUP):
                                to_socket_fileno = None
                                if detect_fileno in self.__client_by_fileno:
                                    to_socket = self.__relay_by_client_fileno.get(detect_fileno)
                                    if to_socket:
                                        to_socket_fileno = to_socket.fileno()
                                        
                                elif detect_fileno in self.__relay_by_fileno:
                                    to_socket = self.__client_by_relay_fileno.get(detect_fileno)
                                    if to_socket:
                                        to_socket_fileno = to_socket.fileno()
                                        
                                if to_socket_fileno:
                                    self.__shutdown_client(to_socket_fileno)
                                
                                if self.__unregister(detect_fileno):
                                    self.__close_client(detect_fileno)
                                    self.__remove_client(detect_fileno)
                                
                            elif not detect_event & (select.EPOLLIN | select.EPOLLOUT):
                                if self.__message_queue:
                                    self.__message_queue.put_nowait({
                                        "type" : "debug",
                                        "message" : f"Unknown Event. {detect_event:#06x}, exist:{detect_fileno in self.__client_by_fileno}"
                                    })
                                
        except Exception as e:
            if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
                self.__message_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"{e}, {traceback.format_exc()}"
                })
            
        if self.__message_queue is not None and isinstance(self.__message_queue, queue.Queue):
            self.__message_queue.put_nowait({
                "type" : "debug",
                "message" : f"Finish Epoll Work"
            })
            


import signal
import time
import iosock
import threading
import socket
import queue
import traceback
import collections
import select
import errno

client_num = 1
send_count = 5
data_sum_count1 = 12
data_sum_count2 = 11

def packing(source_bytes: bytes, starter: bytes = b'', closer: bytes = b'', byteorder:str = 'little') -> bytes:
    bit8_length = 1
    bit16_length = 2
    bit32_length = 4
    bit64_length = 8
    bit8_limit = 256
    bit16_limit = 65536
    bit32_limit = 4294967296
    bit64_limit = 18446744073709551616
    
    limit_bit = -1
    source_length = len(source_bytes)
    if source_length < bit8_limit:
        limit_bit = bit8_length
    elif source_length < bit16_limit:
        limit_bit = bit16_length
    elif source_length < bit32_limit:
        limit_bit = bit32_length
    elif source_length < bit64_limit:
        limit_bit = bit64_length
    
    packet = b''
    if 0<limit_bit:
        source_length_bytes = source_length.to_bytes(limit_bit, byteorder=byteorder)
        length_of_length = len(source_length_bytes)
        length_of_length_bytes = length_of_length.to_bytes(bit8_length, byteorder=byteorder)        
        packet = starter + length_of_length_bytes + source_length_bytes + source_bytes + closer
    return packet

def unpacking(source_bytes: bytes, byteorder: str = 'little') -> bytes:
    bit8_length = 1
    length_of_length = int.from_bytes(source_bytes[:bit8_length], byteorder=byteorder)
    source_length = int.from_bytes(source_bytes[bit8_length:(bit8_length+length_of_length)], byteorder=byteorder)
    start_index = bit8_length+length_of_length
    end_index = bit8_length+length_of_length+source_length
    if len(source_bytes) == end_index:
        return source_bytes[start_index:end_index]
    else:
        return None

send_bytes1 = b'abcdefghijklmnop qrstuvwxyz'
for _ in range(data_sum_count1):
    send_bytes1 += send_bytes1
send_bytes2 = b'abcdefghijklmnop qrstuvwxyz'
for _ in range(int(data_sum_count2)):
    send_bytes2 += send_bytes2
    
print(f"data count:{send_count}+{send_count} send_bytes:{(len(send_bytes1)+len(send_bytes2))*send_count*client_num} send_bytes1:{len(send_bytes1)} send_bytes2:{len(send_bytes2)}")

starter = b'%w$d#'
closer = b'&sa@f#d$'
packed_send_bytes = packing(send_bytes1, starter, closer)
packed_send_bytes_length = len(packed_send_bytes)

packed_send_bytes2 = packing(send_bytes2, starter, closer)
packed_send_bytes_length2 = len(packed_send_bytes2)


kevents = collections.defaultdict(select.kevent)
clients = collections.defaultdict(iosock.DarwinClient)
recv_data = collections.defaultdict(bytes)
recv_data_len = collections.defaultdict(int)
locks = collections.defaultdict(threading.Lock)
import multiprocessing
import ctypes
is_running = multiprocessing.Value(ctypes.c_bool, True)

update_event, update_listener = socket.socketpair()
close_event, close_listener = socket.socketpair()

def kqueueing():
    kq = select.kqueue()
    try:
        # print(type(kevents))
        count = collections.defaultdict(int)
        while is_running.value:
            revents = kq.control(list(kevents.values()), 1000)
            for event in revents:
                if event.flags & select.KQ_EV_ERROR:
                    print(f"[{event.ident}] [{'Bad FD' if event.data == 9 else event.data}] event.flags & select.KQ_EV_ERROR")
                    kevents.pop(event.ident)
                    continue
                    
                if event.ident == close_listener.fileno():
                    print(f"event.ident == detect_close_fd.fileno() {event}")
                    is_running.value = False
                    continue
                
                elif event.ident == update_listener.fileno():
                    print(f"event.ident == update_listener.fileno() in")
                    update_data = update_listener.recv(1024)
                    print(f"event.ident == update_listener.fileno() {update_data}")
                    continue
                    
                elif event.filter == select.KQ_FILTER_READ:
                    if event.flags & select.KQ_EV_EOF:
                        eof_message = f"[{event.ident:3}] event.flags & select.KQ_EV_EOF"
                        if event.ident in kevents:
                            eof_message += f" Pop Event({len(kevents):3}->"
                            e = kevents.pop(event.ident)
                            eof_message += f"{len(kevents):3})"
                        
                        if event.ident in clients:
                            client : iosock.DarwinClient = clients.pop(event.ident)
                            client.close()
                        
                        if event.ident in recv_data:
                            eof_message += f" remain:{len(recv_data[event.ident])}"
                            # if 0<len(recv_data[event.ident]):
                        else:
                            eof_message += f" remain false"

                        # print(eof_message)
                        
                    else:
                        client : iosock.DarwinClient = clients.get(event.ident)
                        if client:
                            data = b''
                            data = client.recv()
                            print(data)
                            
                            if event.ident in recv_data:
                                recv_data[event.ident] += data
                            else:
                                recv_data[event.ident] = data
                                
                            if not event.ident in recv_data_len:
                                recv_data_len[event.ident] = 0
                            
                            recv_data_len[event.ident] += len(data)
                            
                            fileno = event.ident
                            
                            is_start = True
                            is_len = True
                            is_closer = True
                            
                            while is_start and is_len and is_closer:
                                try:
                                    bit8_length = 1
                                    start_index = len(starter)
                                    end_index = len(starter)+bit8_length
                                    is_start = end_index <= len(recv_data[fileno]) and recv_data[fileno][:len(starter)] == starter
                                    length_of_length_bytes = recv_data[fileno][start_index:end_index]
                                    length_of_length = int.from_bytes(length_of_length_bytes, byteorder='little')
                                    
                                    start_index = end_index
                                    end_index = end_index + length_of_length
                                    is_len = end_index <= len(recv_data[fileno])
                                    
                                    length_bytes = recv_data[fileno][start_index:end_index]
                                    source_length = int.from_bytes(length_bytes, byteorder='little')
                                    
                                    start_index = end_index
                                    end_index = end_index+source_length
                                    is_closer = end_index+len(closer) <= len(recv_data[fileno]) and recv_data[fileno][end_index:end_index+len(closer)] == closer
                                except IndexError:
                                    break
                                
                                if is_start and is_len and is_closer:
                                    if event.ident in count:
                                        count[event.ident] += 1
                                    else:
                                        count[event.ident] = 1
                                        
                                    recv_bytes:bytes = recv_data[fileno][:end_index+len(closer)]
                                    recv_data[fileno] = recv_data[fileno][end_index+len(closer):]
                                    # print(f"[{fileno}] {len(recv_bytes)} {recv_bytes[:10]}...{recv_bytes[-10:]}")
                                    try:
                                        if send_count*2 == count[event.ident]:
                                            # print(f"[{fileno:2}] [{count[event.ident]:2}] recv {len(recv_bytes):,} bytes all:{recv_data_len[event.ident]} remain:{len(recv_data[fileno])} End. Try Close")
                                            client : iosock.DarwinClient = clients.get(fileno)
                                            client.shutdown()
                                        # elif send_count*2 < count[event.ident]:
                                            # print(f"[{fileno:2}] [{count[event.ident]:2}] Try Close Warning Count")
                                            # client : iosock.DarwinClient = clients.get(fileno)
                                            # client.shutdown()
                                        # else:
                                        #     print(f"[{fileno:2}] [{count[event.ident]:2}] recv {len(recv_bytes):,} bytes all:{recv_data_len[event.ident]} remain:{len(recv_data[fileno])}")
                                            
                                    except KeyError:
                                        pass
                else:
                    print('else', event)
        
    except Exception as e:
        print(f"recver exception: {e}\n{traceback.format_exc()}")
    kq.close()
    print("finish recv")

def signal_handler(num_recv_signal, frame):
    print(f"Get Signal: {signal.Signals(num_recv_signal).name}")
    is_running.value = False
    try:
        close_event.shutdown(socket.SHUT_RDWR)
    except Exception as e:
        print(e)

from multiprocessing.pool import ThreadPool    

if __name__ == '__main__':
    try:
        pool = ThreadPool(16)
        
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGABRT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        
        kevent_close = select.kevent(close_listener.fileno())
        kevents.update({close_listener.fileno() : kevent_close})
        kevent_update = select.kevent(update_listener.fileno())
        kevents.update({update_listener.fileno() : kevent_update})
        
        kqueueing_thread = threading.Thread(target=kqueueing)
        kqueueing_thread.start()
        
        for _ in range(client_num):
            if not is_running.value:
                break
            client = iosock.DarwinClient()
            client.connect('218.55.118.203', 59012)
            client_fileno = client.fileno()
            # print(f"connect [{client_fileno}]")
            locks[client_fileno] = threading.Lock()
            clients[client_fileno] = client
            kevent = select.kevent(client_fileno)
            kevents.update({client_fileno:kevent})
        
        update_event.send(b'update')
        
        client_filenos = list(clients.keys())
        
        res_pools = []
        for cf in client_filenos:
            def send_data(client_fileno):
                # send_count_index = 0
                send_bytes_len = 0
                for _ in range(send_count):
                    if not is_running.value:
                        break
                    try:
                        try:
                            client = clients.get(client_fileno)
                            if client:
                                client.sendall(packed_send_bytes)
                                send_bytes_len += len(packed_send_bytes)
                                client.sendall(packed_send_bytes2)
                                send_bytes_len += len(packed_send_bytes2)
                            # print(f"[{client_fileno:2}] [{send_count_index:2}] send {len(packed_send_bytes):,} bytes")
                            # send_count_index += 1
                            # print(f"[{client_fileno:2}] [{send_count_index:2}] send {len(packed_send_bytes2):,} bytes")
                            # send_count_index += 1
                        except BrokenPipeError:
                            pass
                        except OSError as e:
                            if e.errno == errno.EBADF:
                                pass
                            else:
                                raise e
                    except Exception as e:
                        print('send',  e, traceback.format_exc())
                    
                return send_bytes_len
            
            res_pool = pool.apply_async(send_data, args=(cf,))
            res_pools.append(res_pool)
            
        for res_pool in res_pools:
            _ = res_pool.get()
            # send_bytes_len = res_pool.get()
            # if send_bytes_len != (send_count * len(packed_send_bytes)) + (send_count * len(packed_send_bytes2)):
            #     print(send_bytes_len)
            
        print("finish send")

        kqueueing_thread.join()
        
        for fd in clients:
            try:
                client:iosock.DarwinClient = clients.get(fd)
                if client:
                    client.close()
            except Exception as e:
                print('main', e)
    except Exception as e:
        print(f"main exception: {e}\n{traceback.format_exc()}")
        


import iosock
import signal
import threading
import time
import math
import errno
import multiprocessing
import ctypes


server = iosock.EpollServer()

starter = b'%w$d#'
closer = b'&sa@f#d$'


def signal_handler(num_recv_signal, frame):
    print(f"\nGet Signal: {signal.Signals(num_recv_signal).name}")
    server.close()
    # test_server.close()
    print("Server Close.")

import collections
count = collections.defaultdict(int)
recv_bytes_by_fileno = collections.defaultdict(bytes)
recvlen = collections.defaultdict(int)

def recv_callback(fileno, recv_bytes) -> list[bytes]:
    send_bytes = []
    if not fileno in count:
        count[fileno] = 0
    count[fileno] += 1

    if fileno in recv_bytes_by_fileno:
        recv_bytes_by_fileno[fileno] += recv_bytes
        recvlen[fileno] += len(recv_bytes)
    
    start_index = -1
    end_index = -1
    
    is_start = True
    is_len = True
    is_closer = True
    
    while is_start and is_len and is_closer:
        try:
            bit8_length = 1
            start_index = len(starter)
            end_index = len(starter)+bit8_length
            is_start = end_index <= len(recv_bytes_by_fileno[fileno]) and recv_bytes_by_fileno[fileno][:len(starter)] == starter
            length_of_length_bytes = recv_bytes_by_fileno[fileno][start_index:end_index]
            length_of_length = int.from_bytes(length_of_length_bytes, byteorder='little')
            
            start_index = end_index
            end_index = end_index + length_of_length
            is_len = end_index <= len(recv_bytes_by_fileno[fileno])
            
            length_bytes = recv_bytes_by_fileno[fileno][start_index:end_index]
            source_length = int.from_bytes(length_bytes, byteorder='little')
            
            start_index = end_index
            end_index = end_index+source_length
            is_closer = end_index+len(closer) <= len(recv_bytes_by_fileno[fileno]) and recv_bytes_by_fileno[fileno][end_index:end_index+len(closer)] == closer
        except IndexError:
            break
        
        if is_start and is_len and is_closer:
            send_bytes.append(recv_bytes_by_fileno[fileno][:end_index+len(closer)])
            
            recv_message_bytes:bytes = recv_bytes_by_fileno[fileno][start_index:end_index]
            # end = time.time()
            
            recv_bytes_replaced = recv_message_bytes.replace(b'abcdefghijklmnop qrstuvwxyz', b'')
            if recv_bytes_replaced != b'':
                recv_bytes_replaced = recv_message_bytes.replace(b'abcdefghijklmnop qrstuvwxyz', b'.')
            
                # time elapsed: {math.floor((end - time_recv_data[fileno])*100000)/100000:.5f}
                text_print = f'[{fileno:3}][{count[fileno]:5}] recv {len(recv_message_bytes):10}/{recvlen[fileno]:10} bytes. over:{len(recv_bytes_by_fileno[fileno]):10}  replaced:{recv_bytes_replaced}'
                print(text_print)
            
            recv_bytes_by_fileno[fileno] = recv_bytes_by_fileno[fileno][end_index+len(closer):]
    
    return send_bytes

is_running = multiprocessing.Value(ctypes.c_bool, True)
threads = []

def recv_threading():
    while is_running.value:
        recv_data = server.recv()
        if recv_data:
            if not recv_data[0] in recv_bytes_by_fileno:
                recv_bytes_by_fileno[recv_data[0]] = b''
                recvlen[recv_data[0]] = 0
        
            send_bytes = recv_callback(recv_data[0], recv_data[1])
            for send_byte in send_bytes:
                server.send(recv_data[0], send_byte)
        else:
            is_running.value = False
            
    print('Finish recv_threading')
    
if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGABRT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # ip = '218.55.118.203'
    # port = 59012
    ip = 'localhost'
    port = 60809

    server.listen(ip, port)
    server.start(count_threads=1)
    for _ in range(5):
        rt = threading.Thread(target=recv_threading)
        rt.start()
        threads.append(rt)
    print("Server Start.")
    
    server.join()
    
    for rt in threads:
        rt.join()
    print("Join Receive Thread.")


import signal
import threading

import iosock

server = iosock.RelayServer()

def check_relay_function(from_socket_fileno:int, to_socket_fileno:int, recv_bytes:bytes) -> bool:
    print(f"{from_socket_fileno:3} -> {to_socket_fileno:3} len:{len(recv_bytes)}")
    return True

def signal_handler(num_recv_signal, frame):
    print(f"\nGet Signal: {signal.Signals(num_recv_signal).name}")
    server.close()
    print("Server Close.")

if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGABRT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    
    server.relay('218.55.118.203', 59012, 'localhost', 60809)
    server.start()
    print("Server Start.")
    
    server.join()
    print("Joined Server.")


import socket
import select
import multiprocessing
import ctypes
import threading
import collections
import queue
import errno
import traceback
from datetime import datetime

class EpollServer():
    def __init__(self) -> None:
        self.__buffer_size = 8196
        self.__is_running = multiprocessing.Value(ctypes.c_bool, False)
        self.__is_debug_mode = multiprocessing.Value(ctypes.c_bool, False)
        self.__running_threads = []
        self.__running_thread_by_tid = collections.defaultdict(threading.Thread)
        
        self.__listener_by_ip_port = collections.defaultdict(socket.socket)
        self.__listener_by_fileno = collections.defaultdict(socket.socket)
        
        self.__client_by_fileno = collections.defaultdict(socket.socket)
        self.__listener_fileno_by_client_fileno = collections.defaultdict(int)
        self.__registered_eventmask_by_fileno = collections.defaultdict(int)
        self.__send_lock_by_fileno = collections.defaultdict(threading.Lock)
        self.__recv_lock_by_fileno = collections.defaultdict(threading.Lock)
        
        self.__send_buffer_by_fileno = collections.defaultdict(bytes)
        self.__send_buffer_lock_by_fileno = collections.defaultdict(threading.Lock)
        
        self.__client_fileno_dict_by_listener_fileno = collections.defaultdict(dict)
        
        self.__recv_queue = queue.Queue()
        
        self.__epoll = None
        
        self.__listener_eventmask = select.EPOLLIN | select.EPOLLPRI | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
        self.__recv_eventmask = select.EPOLLIN  | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
        self.__send_recv_eventmask = select.EPOLLIN | select.EPOLLOUT | select.EPOLLHUP | select.EPOLLRDHUP
        self.__closer_eventmask = select.EPOLLIN | select.EPOLLPRI | select.EPOLLHUP | select.EPOLLRDHUP | select.EPOLLET
    
    def listen(self, ip:str, port:int, backlog:int = 5):
        listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1) # Nagle's
        
        # increase buffer size
        recv_buf_size = listener.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
        send_buf_size = listener.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, recv_buf_size*2)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, send_buf_size*2)
        
        listener.setblocking(False)
        listener.bind((ip, port))
        listener.listen(backlog)
        
        listener_fileno = listener.fileno()
        
        self.__listener_by_ip_port.update({f"{ip}:{port}":listener})
        self.__listener_by_fileno.update({listener_fileno : listener})
        
        self.__client_fileno_dict_by_listener_fileno.update({listener_fileno : {}})
        
        if self.__epoll and not self.__epoll.closed:
            # After 'start()'
            self.__epoll.register(listener_fileno, self.__listener_eventmask)
            self.__registered_eventmask_by_fileno.update({listener_fileno : self.__listener_eventmask})

    def unlisten(self, ip:str, port:int):
        listener = self.__listener_by_ip_port.get(f"{ip}:{port}")
        if listener:
            listener.shutdown(socket.SHUT_RDWR)
    
    def start(self, count_threads:int=1, is_debug_mode:bool = False):
        self.__is_running.value = True
        self.__is_debug_mode.value = is_debug_mode
        
        self.__epoll = select.epoll()
        self.__close_event, self.__close_event_listener = socket.socketpair()
        self.__epoll.register(self.__close_event_listener, self.__closer_eventmask)
        
        for _ in range(count_threads):
            et = threading.Thread(target=self.__epoll_thread_function)
            et.start()
            self.__running_threads.append(et)
            self.__running_thread_by_tid[et.ident] = et
            
        for fileno in self.__listener_by_fileno:
            if fileno in self.__registered_eventmask_by_fileno:
                if self.__registered_eventmask_by_fileno[fileno] != self.__listener_eventmask:
                    self.__epoll.modify(fileno, self.__listener_eventmask)
            else:
                self.__epoll.register(fileno, self.__listener_eventmask)
                self.__registered_eventmask_by_fileno.update({fileno : self.__listener_eventmask})

    def recv(self) -> dict:
        '''
        Return
        -
        dict\n
        return {} if error or close \n
        dict['type'] : 'accept', 'recv', 'debug', 'close_client'\n
        if dict['type'] == 'accept':
            dict['fileno'] (int) : socket fileno
        elif dict['type'] == 'recv':
            dict['fileno'] (int) : socket fileno
            dict['bytes'] (bytes) : receive bytes
        elif dict['type'] == 'debug':
            dict['message'] (str) : debug message
        elif dict['type'] == 'close_client':
            dict['fileno'] (int) : socket fileno
            
        '''
        if self.__is_running.value:
            recv_data = self.__recv_queue.get()
            if recv_data:
                return recv_data
            else:
                #for multithreading
                self.__is_running.value = False
                self.__recv_queue.put_nowait(None)
                return {}
        else:
            #for multithreading
            self.__recv_queue.put_nowait(None)
            return {}
    
    def send(self, socket_fileno:int, data:bytes):
        if data is not None and data != b'':
            try:
                send_buffer_lock = self.__send_buffer_lock_by_fileno.get(socket_fileno)
                if send_buffer_lock:
                    with send_buffer_lock:
                        send_buffer = self.__send_buffer_by_fileno.get(socket_fileno)
                        if send_buffer is not None:
                            self.__send_buffer_by_fileno.update({socket_fileno : send_buffer + data})
                    
                send_lock = self.__send_lock_by_fileno.get(socket_fileno)
                if send_lock:
                    with send_lock:
                        registered_eventmask = self.__registered_eventmask_by_fileno.get(socket_fileno)
                        if registered_eventmask is not None and registered_eventmask != self.__send_recv_eventmask:
                            self.__registered_eventmask_by_fileno.update({socket_fileno : self.__send_recv_eventmask})
                            if isinstance(self.__epoll, select.epoll):
                                self.__epoll.modify(socket_fileno, self.__send_recv_eventmask)
                
            except KeyError:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send KeyError.\n{traceback.format_exc()}"
                    })
                
            except FileNotFoundError:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send FileNotFoundError.\n{traceback.format_exc()}"
                    })
                    
            except PermissionError:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{socket_fileno}] send PermissionError.\n{traceback.format_exc()}"
                    })
                
            except OSError as e:
                if e.errno == errno.EBADF:
                    pass
                else:
                    raise e
    
    def join(self):
        for t in self.__running_threads:
            t:threading.Thread = t
            t.join()
                
    def close(self):
        self.__is_running.value = False
        self.__shutdown_listeners()
        
        for _ in self.__running_threads:
            self.__close_event.send(b'close')
            tid_bytes = self.__close_event.recv(32)
            tid = int.from_bytes(tid_bytes, byteorder='big')
            self.__running_thread_by_tid[tid].join()
            
        self.__recv_queue.put_nowait(None)
    
    def __shutdown_listeners(self):
        fileno_list = list(self.__listener_by_fileno.keys())
        for fileno in fileno_list:
            self.__shutdown_listener(fileno)
            
    def __shutdown_listener(self, listener_fileno:int):
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            listener.shutdown(socket.SHUT_RDWR)
        
    def __close_listener(self, listener_fileno:int):
        try:
            if isinstance(self.__epoll, select.epoll):
                self.__epoll.unregister(listener_fileno)
        except FileNotFoundError:
            pass
        except OSError as e:
            if e.errno == errno.EBADF:
                pass   
            else:
                raise e
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            listener.close()
            
    def __remove_listener(self, listener_fileno:int):
        try:
            _ = self.__listener_by_fileno.pop(listener_fileno)
        except KeyError:
            pass
        # self.__listener_by_ip_port = collections.defaultdict(socket.socket)
        
    def __unregister(self, socket_fileno:int) -> bool:
        result = False
        try:
            _ = self.__registered_eventmask_by_fileno.pop(socket_fileno)
        except KeyError:
            pass
        
        try:
            if isinstance(self.__epoll, select.epoll):
                self.__epoll.unregister(socket_fileno)
                result = True
        
        except FileNotFoundError:
            result = True
            
        except OSError as e:
            if e.errno == errno.EBADF:
                result = True
            else:
                raise e
        return result
        
    def __shutdown_clients_by_listener(self, listener_fileno:int):
        client_fileno_dict = self.__client_fileno_dict_by_listener_fileno.get(listener_fileno)
        if client_fileno_dict:
            client_fileno_list = list(client_fileno_dict.keys())
            for client_fileno in client_fileno_list:
                self.shutdown_client(client_fileno)
        
    def shutdown_client(self, client_fileno:int):
        client_socket = self.__client_by_fileno.get(client_fileno)
        if client_socket:
            try:
                client_socket.shutdown(socket.SHUT_RDWR)
            except ConnectionResetError:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{client_fileno}] shutdown_client ConnectionResetError.\n{traceback.format_exc()}"
                    })
            
            except BrokenPipeError:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{client_fileno}] shutdown_client BrokenPipeError.\n{traceback.format_exc()}"
                    })
            
            except OSError as e:
                if e.errno == errno.ENOTCONN: # errno 107
                    pass
                else:
                    raise e
        
    def __close_client(self, client_fileno:int):
        client_socket = self.__client_by_fileno.get(client_fileno)
        if client_socket:
            client_socket.close()
        
            self.__recv_queue.put_nowait({
                "type" : "close_client",
                "fileno" : client_fileno
            })
    
            
    def __remove_client(self, client_fileno:int):
        try:
            try: _ = self.__send_lock_by_fileno.pop(client_fileno)
            except KeyError: pass
            try: _ = self.__recv_lock_by_fileno.pop(client_fileno)
            except KeyError: pass
            try: _ = self.__client_by_fileno.pop(client_fileno)
            except KeyError: pass
            
            len_remain_send_buffer = 0
            try: 
                remain_send_buffer = self.__send_buffer_by_fileno.pop(client_fileno)
                len_remain_send_buffer = len(remain_send_buffer)
            except KeyError: pass
            
            try:
                listener_fileno = self.__listener_fileno_by_client_fileno.pop(client_fileno)
                _ = self.__client_fileno_dict_by_listener_fileno[listener_fileno].pop(client_fileno)
            except KeyError: pass
            
            if 0 < len_remain_send_buffer:
                if self.__is_debug_mode.value:
                    self.__recv_queue.put_nowait({
                        "type" : "debug",
                        "message" : f"[{client_fileno}] Removed. But send buffer remain:{len_remain_send_buffer} bytes."
                    })
        except Exception as e:
            if self.__is_debug_mode.value:
                self.__recv_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{client_fileno}] Remove Client Exception:{e}.\n{traceback.format_exc()}"
                })
        
        
        
    
    def __epoll_accept(self, listener_fileno:int):
        listener = self.__listener_by_fileno.get(listener_fileno)
        if listener:
            try:
                client_socket, address = listener.accept()
                client_socket_fileno = client_socket.fileno()
                client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                client_socket.setblocking(False)
                
                self.__client_by_fileno.update({client_socket_fileno : client_socket})
                self.__send_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__recv_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__send_buffer_lock_by_fileno.update({client_socket_fileno : threading.Lock()})
                self.__send_buffer_by_fileno.update({client_socket_fileno : b''})
                if not listener_fileno in self.__client_fileno_dict_by_listener_fileno:
                    self.__client_fileno_dict_by_listener_fileno.update({listener_fileno : {}})
                self.__client_fileno_dict_by_listener_fileno[listener_fileno][client_socket_fileno] = True
                self.__listener_fileno_by_client_fileno.update({client_socket_fileno : listener_fileno})
                
                self.__registered_eventmask_by_fileno[client_socket_fileno] = self.__recv_eventmask
                if isinstance(self.__epoll, select.epoll):
                    self.__epoll.register(client_socket, self.__recv_eventmask)
                
                self.__recv_queue.put_nowait({
                    "type" : "accept",
                    "fileno" : client_socket_fileno
                })
        
            except BlockingIOError as e:
                if e.errno == socket.EAGAIN:
                    pass
                else:
                    raise e
    
    def __epoll_recv(self, client_fileno:int):
        is_connect = True
        recv_lock = self.__recv_lock_by_fileno.get(client_fileno)
        if recv_lock:
            with recv_lock:
                recv_bytes = b''
                client_socket = self.__client_by_fileno.get(client_fileno)
                if client_socket:
                    is_eagain = False
                    try:
                        temp_recv_bytes = client_socket.recv(self.__buffer_size)
                        if temp_recv_bytes == None or temp_recv_bytes == -1 or temp_recv_bytes == b'':
                            if self.__is_debug_mode.value:
                                self.__recv_queue.put_nowait({
                                    "type" : "debug",
                                    "message" : f"[{client_fileno}] recv break :'{temp_recv_bytes}'"
                                })
                        
                            is_connect = False
                        else:
                            recv_bytes += temp_recv_bytes
                            
                    except ConnectionError as e:
                        pass
                    
                    except OSError as e:
                        if e.errno == socket.EAGAIN:
                            is_eagain = True
                        elif e.errno == errno.EBADF:
                            is_connect = False
                        else:
                            raise e

                    if not is_eagain and is_connect:
                        try:
                            if isinstance(self.__epoll, select.epoll):
                                self.__epoll.modify(client_fileno, self.__registered_eventmask_by_fileno[client_fileno])
                        except FileNotFoundError:
                            pass
                        except OSError as e:
                            if e.errno == errno.EBADF:
                                pass

                    if recv_bytes:
                        self.__recv_queue.put_nowait({
                            "type" : "recv",
                            "fileno" : client_fileno,
                            "bytes" : recv_bytes
                        })
                        
        return is_connect
    
    def __epoll_send(self, client_fileno:int):
        try:
            is_connect = True
            client_socket = self.__client_by_fileno.get(client_fileno)
            send_lock = self.__send_lock_by_fileno.get(client_fileno)
            send_buffer_lock = self.__send_buffer_lock_by_fileno.get(client_fileno)
            
            if send_lock is not None and client_socket is not None:
                with send_lock:
                    send_buffer = b''
                    if send_buffer_lock:
                        with send_buffer_lock:
                            send_buffer = self.__send_buffer_by_fileno.get(client_fileno)
                            self.__send_buffer_by_fileno.update({client_fileno: b''})
                    else:
                        is_connect = False
        
                    if send_buffer is not None:
                        if send_buffer != b'':
                            try:
                                sent_length = client_socket.send(send_buffer)
                                if 0 < sent_length:
                                    if send_buffer_lock:
                                        with send_buffer_lock:
                                            next_send_buffer = self.__send_buffer_by_fileno.get(client_fileno)
                                            if next_send_buffer is None:
                                                next_send_buffer = b''
                                            self.__send_buffer_by_fileno.update({client_fileno : send_buffer[sent_length:] + next_send_buffer})
                        
                            except ConnectionError as e:
                                pass
                            
                            except BlockingIOError as e:
                                if e.errno == socket.EAGAIN:
                                    pass
                                else:
                                    raise e
                                
                            except OSError as e:
                                if e.errno == errno.EBADF:
                                    is_connect = False
                                    
                                else:
                                    raise e
                    else:
                        is_connect = False

            else:
                is_connect = False
        
                
            try:
                send_buffer_lock = self.__send_buffer_lock_by_fileno.get(client_fileno)
                if send_buffer_lock:
                    remain_buffer = b''
                    with send_buffer_lock:
                        remain_buffer = self.__send_buffer_by_fileno.get(client_fileno)
                    registered_eventmask = self.__registered_eventmask_by_fileno.get(client_fileno)
                    
                    if (remain_buffer is not None and registered_eventmask is not None) and\
                        remain_buffer == b'' and\
                        registered_eventmask != self.__recv_eventmask:
                        self.__registered_eventmask_by_fileno.update({client_fileno : self.__recv_eventmask})
                        if isinstance(self.__epoll, select.epoll):
                            self.__epoll.modify(client_fileno, self.__recv_eventmask)
            except OSError as e:
                if e.errno == errno.EBADF:
                    is_connect = False
                        
        except Exception as e:
            if self.__is_debug_mode.value:
                self.__recv_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"[{client_fileno:3}] __epoll_send Exception:{e}\n{traceback.format_exc()}"
                })
            is_connect = False
            
        return is_connect
                
    def __epoll_thread_function(self):
        __is_running = True
        tid = threading.get_ident()
        try:
            if isinstance(self.__epoll, select.epoll):
                while __is_running:
                    events = self.__epoll.poll()
                    for detect_fileno, detect_event in events:
                        if detect_event & select.EPOLLPRI:
                            pass
                        if detect_fileno == self.__close_event_listener.fileno():
                            self.__close_event_listener.send(tid.to_bytes(32, 'big'))
                            __is_running = False
                            
                        elif detect_fileno in self.__listener_by_fileno:
                            if detect_event & (select.EPOLLHUP | select.EPOLLRDHUP):
                                self.__shutdown_clients_by_listener(detect_fileno)
                                if self.__unregister(detect_fileno):
                                    self.__close_listener(detect_fileno)
                                    self.__remove_listener(detect_fileno)
                                
                            elif detect_event & select.EPOLLIN:
                                self.__epoll_accept(detect_fileno)
                            
                            else:
                                if self.__is_debug_mode.value:
                                    self.__recv_queue.put_nowait({
                                        "type" : "debug",
                                        "message" : f"listener got unknown event : {detect_event:#06x}"
                                    })
                            
                        elif detect_fileno in self.__client_by_fileno:
                            unregistered = False
                            if detect_event & select.EPOLLOUT:
                                if self.__epoll_send(detect_fileno) == False:
                                    if self.__is_debug_mode.value:
                                        self.__recv_queue.put_nowait({
                                            "type" : "debug",
                                            "message" : f"self.__epoll_send False"
                                        })
                                
                                    if self.__unregister(detect_fileno):
                                        unregistered = True
                                        self.__close_client(detect_fileno)
                                        self.__remove_client(detect_fileno)
                            
                            if detect_event & select.EPOLLIN:
                                if self.__epoll_recv(detect_fileno) == False:
                                    if self.__is_debug_mode.value:
                                        self.__recv_queue.put_nowait({
                                            "type" : "debug",
                                            "message" : f"self.__epoll_recv False"
                                        })
                                
                                    if self.__unregister(detect_fileno):
                                        unregistered = True
                                        self.__close_client(detect_fileno)
                                        self.__remove_client(detect_fileno)
                            
                            if detect_event & (select.EPOLLHUP | select.EPOLLRDHUP):
                                if self.__is_debug_mode.value:
                                    self.__recv_queue.put_nowait({
                                        "type" : "debug",
                                        "message" : f"[{detect_fileno}] detect_event & (select.EPOLLHUP | select.EPOLLRDHUP)"
                                    })
                            
                                if not unregistered:
                                    if self.__unregister(detect_fileno):
                                        self.__close_client(detect_fileno)
                                        self.__remove_client(detect_fileno)
                                
                        else:
                            if self.__is_debug_mode.value:
                                self.__recv_queue.put_nowait({
                                    "type" : "debug",
                                    "message" : f"[{detect_fileno:3}] Unknown Fileno. {detect_event:#06x}, exist:{detect_fileno in self.__client_by_fileno}"
                                })
                        
        except Exception as e:
            if self.__is_debug_mode.value:
                self.__recv_queue.put_nowait({
                    "type" : "debug",
                    "message" : f"Exception in epoll_thread_function.\n{e}.\n{traceback.format_exc()}"
                })